@@ -704,6 +704,7 @@ static void save_history_based_model(struct starpu_perfmodel *model)
 
 static void _starpu_dump_registered_models(void)
 {
+#ifndef STARPU_SIMGRID
 	STARPU_PTHREAD_RWLOCK_WRLOCK(&registered_models_rwlock);
 
 	struct _starpu_perfmodel_list *node;
@@ -718,6 +719,7 @@ static void _starpu_dump_registered_models(void)
 	}
 
 	STARPU_PTHREAD_RWLOCK_UNLOCK(&registered_models_rwlock);
+#endif
 }
 
 void _starpu_initialize_registered_performance_models(void)
@@ -1212,17 +1214,19 @@ void _starpu_update_perfmodel_history(struct _starpu_job *j, struct starpu_perfm
 				STARPU_HG_DISABLE_CHECKING(entry->nsample);
 				STARPU_HG_DISABLE_CHECKING(entry->mean);
 
-				entry->mean = measured;
-				entry->sum = measured;
+				/* Do not take the first measurement into account, it is very often quite bogus */
+				/* TODO: it'd be good to use a better estimation heuristic, like the median, or latest n values, etc. */
+				entry->mean = 0;
+				entry->sum = 0;
 
 				entry->deviation = 0.0;
-				entry->sum2 = measured*measured;
+				entry->sum2 = 0;
 
 				entry->size = _starpu_job_get_data_size(model, arch, nimpl, j);
 				entry->flops = j->task->flops;
 
 				entry->footprint = key;
-				entry->nsample = 1;
+				entry->nsample = 0;
 
 				insert_history_entry(entry, list, &per_arch_model->history);
 			}
