@@ -1,6 +1,6 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2009-2013  Université de Bordeaux 1
+ * Copyright (C) 2009-2014  Université de Bordeaux 1
  * Copyright (C) 2010, 2011, 2012, 2013  Centre National de la Recherche Scientifique
  *
  * StarPU is free software; you can redistribute it and/or modify
@@ -38,6 +38,8 @@ static struct handle_entry *registered_handles;
 static struct _starpu_spinlock    registered_handles_lock;
 static int _data_interface_number = STARPU_MAX_INTERFACE_ID;
 
+static void _starpu_data_unregister(starpu_data_handle_t handle, unsigned coherent, unsigned nowait);
+
 void _starpu_data_interface_init(void)
 {
 	_starpu_spin_init(&registered_handles_lock);
@@ -152,6 +154,7 @@ static void _starpu_register_new_data(starpu_data_handle_t handle,
 
 	handle->reduction_refcnt = 0;
 	handle->reduction_req_list = _starpu_data_requester_list_new();
+	handle->reduction_tmp_handles = NULL;
 
 #ifdef STARPU_USE_FXT
 	handle->last_submitted_ghost_writer_id_is_valid = 0;
@@ -185,12 +188,14 @@ static void _starpu_register_new_data(starpu_data_handle_t handle,
 			replicate->state = STARPU_OWNER;
 			replicate->allocated = 1;
 			replicate->automatically_allocated = 0;
+			replicate->initialized = 1;
 		}
 		else
 		{
 			/* the value is not available here yet */
 			replicate->state = STARPU_INVALID;
 			replicate->allocated = 0;
+			replicate->initialized = 0;
 		}
 	}
 
@@ -205,10 +210,10 @@ static void _starpu_register_new_data(starpu_data_handle_t handle,
 		replicate->state = STARPU_INVALID;
 		replicate->refcnt = 0;
 		replicate->handle = handle;
+		replicate->requested = 0;
 
 		for (node = 0; node < STARPU_MAXNODES; node++)
 		{
-			replicate->requested[node] = 0;
 			replicate->request[node] = NULL;
 		}
 
@@ -334,42 +339,42 @@ int starpu_data_get_rank(starpu_data_handle_t handle)
 	return handle->rank;
 }
 
-int starpu_data_set_rank(starpu_data_handle_t handle, int rank)
+int _starpu_data_set_rank(starpu_data_handle_t handle, int rank)
 {
         handle->rank = rank;
         return 0;
 }
 
+int starpu_data_set_rank(starpu_data_handle_t handle, int rank)
+{
+	_STARPU_DISP("Warning: You should call starpu_mpi_data_register which will insure MPI cache will be cleared when unregistering the data\n");
+	return _starpu_data_set_rank(handle, rank);
+}
+
 int starpu_data_get_tag(starpu_data_handle_t handle)
 {
 	return handle->tag;
 }
 
-int starpu_data_set_tag(starpu_data_handle_t handle, int tag)
+int _starpu_data_set_tag(starpu_data_handle_t handle, int tag)
 {
         handle->tag = tag;
         return 0;
 }
 
+int starpu_data_set_tag(starpu_data_handle_t handle, int tag)
+{
+	_STARPU_DISP("Warning: You should call starpu_mpi_data_register which will insure MPI cache will be cleared when unregistering the data\n");
+	return _starpu_data_set_tag(handle, tag);
+}
+
 /*
  * Stop monitoring a piece of data
  */
 
-void _starpu_data_free_interfaces(starpu_data_handle_t handle)
+void _starpu_data_unregister_ram_pointer(starpu_data_handle_t handle)
 {
-	const void *ram_ptr;
-	unsigned node;
-	unsigned worker;
-	unsigned nworkers = starpu_worker_get_count();
-
-	ram_ptr = starpu_data_handle_to_pointer(handle, 0);
-
-	for (node = 0; node < STARPU_MAXNODES; node++)
-		free(handle->per_node[node].data_interface);
-
-	for (worker = 0; worker < nworkers; worker++)
-		free(handle->per_worker[worker].data_interface);
-
+	const void *ram_ptr = starpu_data_handle_to_pointer(handle, 0);
 	if (ram_ptr != NULL)
 	{
 		/* Remove the PTR -> HANDLE mapping.  If a mapping from PTR
@@ -388,6 +393,19 @@ void _starpu_data_free_interfaces(starpu_data_handle_t handle)
 	}
 }
 
+void _starpu_data_free_interfaces(starpu_data_handle_t handle)
+{
+	unsigned node;
+	unsigned worker;
+	unsigned nworkers = starpu_worker_get_count();
+
+	for (node = 0; node < STARPU_MAXNODES; node++)
+		free(handle->per_node[node].data_interface);
+
+	for (worker = 0; worker < nworkers; worker++)
+		free(handle->per_worker[worker].data_interface);
+}
+
 struct _starpu_unregister_callback_arg
 {
 	unsigned memory_node;
@@ -417,7 +435,7 @@ int _starpu_data_check_not_busy(starpu_data_handle_t handle)
 	if (handle->lazy_unregister && handle->busy_count == 0)
 	{
 		_starpu_spin_unlock(&handle->header_lock);
-		starpu_data_unregister_no_coherency(handle);
+		_starpu_data_unregister(handle, 0, 1);
 		/* Warning: in case we unregister the handle, we must be sure
 		 * that the caller will not try to unlock the header after
 		 * !*/
@@ -449,16 +467,27 @@ static void _starpu_data_unregister_fetch_data_callback(void *_arg)
 }
 
 /* Unregister the data handle, perhaps we don't need to update the home_node
- * (in that case coherent is set to 0) */
-static void _starpu_data_unregister(starpu_data_handle_t handle, unsigned coherent)
+ * (in that case coherent is set to 0)
+ * nowait is for internal use when we already know for sure that we won't have to wait.
+ */
+static void _starpu_data_unregister(starpu_data_handle_t handle, unsigned coherent, unsigned nowait)
 {
 	STARPU_ASSERT(handle);
 	STARPU_ASSERT_MSG(handle->nchildren == 0, "data %p needs to be unpartitioned before unregistration", handle);
+	STARPU_ASSERT(!(nowait && handle->busy_count != 0));
 
-	if (coherent)
+	int sequential_consistency = handle->sequential_consistency;
+	if (sequential_consistency && !nowait)
 	{
+		STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "starpu_data_unregister must not be called from a task or callback, perhaps you can use starpu_data_unregister_submit instead");
+
 		/* If sequential consistency is enabled, wait until data is available */
-		_starpu_data_wait_until_available(handle, STARPU_RW);
+		_starpu_data_wait_until_available(handle, STARPU_RW, "starpu_data_unregister");
+	}
+
+	if (coherent && !nowait)
+	{
+		STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "starpu_data_unregister must not be called from a task or callback, perhaps you can use starpu_data_unregister_submit instead");
 
 		/* Fetch data in the home of the data to ensure we have a valid copy
 		 * where we registered it */
@@ -574,7 +603,7 @@ static void _starpu_data_unregister(starpu_data_handle_t handle, unsigned cohere
 
 	size_t size = _starpu_data_get_size(handle);
 
-	_starpu_data_free_interfaces(handle);
+	_starpu_data_unregister_ram_pointer(handle);
 
 	/* Destroy the data now */
 	unsigned node;
@@ -594,6 +623,7 @@ static void _starpu_data_unregister(starpu_data_handle_t handle, unsigned cohere
 		if (local->allocated && local->automatically_allocated)
 			_starpu_request_mem_chunk_removal(handle, local, starpu_worker_get_memory_node(worker), size);
 	}
+	_starpu_data_free_interfaces(handle);
 
 	_starpu_memory_stats_free(handle);
 	_starpu_data_requester_list_delete(handle->req_list);
@@ -612,21 +642,50 @@ static void _starpu_data_unregister(starpu_data_handle_t handle, unsigned cohere
 void starpu_data_unregister(starpu_data_handle_t handle)
 {
 	STARPU_ASSERT_MSG(!handle->lazy_unregister, "data %p can not be unregistered twice", handle);
-	_starpu_data_unregister(handle, 1);
+
+	if (handle->unregister_hook)
+	{
+		handle->unregister_hook(handle);
+	}
+
+	_starpu_data_unregister(handle, 1, 0);
 }
 
 void starpu_data_unregister_no_coherency(starpu_data_handle_t handle)
 {
-	_starpu_data_unregister(handle, 0);
+	if (handle->unregister_hook)
+	{
+		handle->unregister_hook(handle);
+	}
+
+	_starpu_data_unregister(handle, 0, 0);
+}
+
+static void _starpu_data_unregister_submit_cb(void *arg)
+{
+	starpu_data_handle_t handle = arg;
+
+	_starpu_spin_lock(&handle->header_lock);
+	handle->lazy_unregister = 1;
+	/* The handle should be busy since we are working on it.
+         * when data_acquire releases the handle, it will be destroyed by _starpu_data_check_not_busy */
+	STARPU_ASSERT(handle->busy_count);
+        _starpu_spin_unlock(&handle->header_lock);
+
+	starpu_data_release_on_node(handle, -1);
 }
 
 void starpu_data_unregister_submit(starpu_data_handle_t handle)
 {
-	_starpu_spin_lock(&handle->header_lock);
 	STARPU_ASSERT_MSG(!handle->lazy_unregister, "data %p can not be unregistered twice", handle);
-	handle->lazy_unregister = 1;
-	_starpu_spin_unlock(&handle->header_lock);
-	_starpu_data_unregister(handle, 0);
+
+	if (handle->unregister_hook)
+	{
+		handle->unregister_hook(handle);
+	}
+
+	/* Wait for all task dependencies on this handle before putting it for free */
+	starpu_data_acquire_on_node_cb(handle, -1, STARPU_RW, _starpu_data_unregister_submit_cb, handle);
 }
 
 static void _starpu_data_invalidate(void *data)
@@ -641,8 +700,13 @@ static void _starpu_data_invalidate(void *data)
 		struct _starpu_data_replicate *local = &handle->per_node[node];
 
 		if (local->mc && local->allocated && local->automatically_allocated)
+		{
+			if (node == 0)
+				_starpu_data_unregister_ram_pointer(handle);
+
 			/* free the data copy in a lazy fashion */
 			_starpu_request_mem_chunk_removal(handle, local, node, size);
+		}
 
 		local->state = STARPU_INVALID;
 	}
@@ -662,14 +726,14 @@ static void _starpu_data_invalidate(void *data)
 
 	_starpu_spin_unlock(&handle->header_lock);
 
-	starpu_data_release(handle);
+	starpu_data_release_on_node(handle, -1);
 }
 
 void starpu_data_invalidate(starpu_data_handle_t handle)
 {
 	STARPU_ASSERT(handle);
 
-	starpu_data_acquire(handle, STARPU_W);
+	starpu_data_acquire_on_node(handle, -1, STARPU_W);
 
 	_starpu_data_invalidate(handle);
 }
@@ -678,7 +742,7 @@ void starpu_data_invalidate_submit(starpu_data_handle_t handle)
 {
 	STARPU_ASSERT(handle);
 
-	starpu_data_acquire_cb(handle, STARPU_W, _starpu_data_invalidate, handle);
+	starpu_data_acquire_on_node_cb(handle, -1, STARPU_W, _starpu_data_invalidate, handle);
 }
 
 enum starpu_data_interface_id starpu_data_get_interface_id(starpu_data_handle_t handle)
